/**
 * @file lin.c
 *
 * @brief This file contains LIN bus routines for the USBlini firmware
 *        project.
 *
 * @author Thomas Fischl (https://www.fischl.de/contact.html)
 * @copyright (c) 2021 Thomas Fischl
 * 
 * This file is part of the USBlini firmware project and is copyrighted by the
 * authors listed above. It is free for private or educational non-commercial use
 * (for a commercial license please contact the authors). It may not be
 * redistributed without the prior written consent of the authors.
 *
 * It is provided "as is" without warranty of any kind, either express or implied,
 * including without limitation any implied warranties of condition, uninterrupted
 * use, merchantability, fitness for a particular purpose, or non-infringement. In
 * no event shall the authors be liable for any direct or indirect damages arising
 * in any way out of the use of it.
 *
 */

#include "usblini.h"
#include "led.h"
#include "usb.h"
#include "clock.h"
#include "linslave.h"
#include "linmaster.h"
#include "statusreport.h"
#include "lin.h"


#define LIN_SYNC 0x55

#define LIN_PIN_CS LATCbits.LATC2

uint8_t timeoutpreset;
volatile uint8_t timeoutticker = 0;
volatile uint8_t framereceived;
uint16_t configuredSPBRG;
uint8_t autobaud_suspended = 0;
uint8_t autobaud_activated = 0;
struct frameinfo finfo;

uint8_t sendstate = SENDSTATE_IDLE;
uint8_t sendpid;
uint8_t senddata[SENDBUFFER_SIZE];
uint8_t senddatalen;
uint8_t senddatapos;
uint8_t sendsource = REPORT_SOURCE_COMMON;

#define ERROR_REPORT_SIZE 16
uint8_t errorreport[ERROR_REPORT_SIZE];

volatile uint8_t receive_state = RECEIVESTATE_IDLE;


/**
 * @brief Initialize LIN
 */
void lin_init(void) {

    // LIN uses LSB first -> perfect, matches with EUSART
    
    // set baudrate (16bit generator)
    BAUDCON = 0x08;
    lin_setBaudrate(LIN_BAUDRATE_DEFAULT);

    // enable transmitter
    RCSTA = 0x00; // clear error flags
    TXSTA = 0x24;
    RCSTA = 0x90;
    RCIE = 1;
    
    // activate transceiver
    LIN_PIN_CS = 1;
    
    uint8_t i;
    for (i = 0; i < ERROR_REPORT_SIZE; i++) {
        errorreport[i] = 0;
    }        
}

/**
 * @brief Set given baudrate
 * @param baudrate Baudrate in Hz
 */
void lin_setBaudrate(uint16_t baudrate) {

    if (baudrate < LIN_BAUDRATE_MIN) baudrate = LIN_BAUDRATE_MIN;
    if (baudrate > LIN_BAUDRATE_MAX) baudrate = LIN_BAUDRATE_MAX;
        
    SPBRG = (uint16_t)((_XTAL_FREQ / 4) / baudrate - 1);
    configuredSPBRG = SPBRG;
    
    timeoutpreset = (uint8_t)(200000 / baudrate); // 100 kHz / 8 / baudrate * 16   
}

/**
 * @brief Set configured baudrate register
 */
void lin_setConfiguredSPBRG(void) {
    SPBRG = configuredSPBRG;    
}

/**
 * @brief Set auto baudrate (slave mode)
 * @param active Autobaud active (0=disabled, 1=enabled)
 */
void lin_setAutobaud(uint8_t active) {
    autobaud_activated = active & 1;
}

/**
 * @brief Do module processing
 */
void lin_processSend(void) {
    
    if (sendstate == SENDSTATE_IDLE) return;
    
    switch (sendstate) {
        
        case SENDSTATE_BREAKSYNC:
            
            if (TXSTAbits.TRMT) {
                // break and sync sent
                
                // now send pid
                TXREG = sendpid;
                sendstate = SENDSTATE_PID;
            }
                
            break;
            
        case SENDSTATE_PID:
            
            if (TXSTAbits.TRMT) {
                // pid sent
                
                // check if we received echo. If not, something is wrong
                if (lin_getReceiveState() == RECEIVESTATE_IDLE) {
                    
                    if (sendsource == REPORT_SOURCE_USER) {
                        errorreport[0] = sendsource | REPORT_TYPE_ERROR;
                        errorreport[1] = ERRORID_NO_ECHO;
                        usb_ep1_put(errorreport, ERROR_REPORT_SIZE);
                    }
                    
                    statusreport_setErrorflag(ERRORFLAG_MISSING_ECHO);
                    
                    sendstate = SENDSTATE_IDLE;
                } else {
                    // now send data                
                    sendstate = SENDSTATE_DATA;
                    finfo.reportid = sendsource | REPORT_TYPE_FRAME;
                }
                
            }
            
            break;
            
        case SENDSTATE_DATA:
            
            if (TXSTAbits.TRMT) {
                
                if (senddatapos >= senddatalen) {
                    // we have finished sending
                    lin_autobaudSuspend(0);
                    sendstate = SENDSTATE_IDLE;
                } else {
                    
                    TXREG = senddata[senddatapos];
                    senddatapos++;                                                            
                }                
            }
            
            break;        
    }    
}

/**
 * @brief Send header of LIN message
 * @param id Identifier of LIN message
 * @param source Source (see REPORT_SOURCE_*)
 */
void lin_sendHeader(uint8_t id, uint8_t source) {

    sendpid = lin_getPID(id);
    senddatapos = 0;
    senddatalen = 0;
    sendsource = source;
   
    lin_autobaudSuspend(1);
    lin_setConfiguredSPBRG();
    
    sendstate = SENDSTATE_BREAKSYNC;
    
    TXSTAbits.SENDB = 1;
    TXREG = 0x00;
    TXREG = LIN_SYNC;
}

/**
 * @brief Send data with header of LIN message
 * @param id Identifier of LIN message
 * @param data Data of LIN message
 * @param length Data length
 * @param source Source of message (see REPORT_SOURCE_*)
 */
void lin_sendDataWithHeader(uint8_t id, uint8_t * data, uint8_t length, uint8_t source) {
    
    uint8_t i;
    
    sendpid = lin_getPID(id);
    sendsource = source;
    senddatapos = 0;
    senddatalen = length;
    for (i = 0; i < length; i++) {
        senddata[i] = data[i];
    }   

    lin_autobaudSuspend(1);
    lin_setConfiguredSPBRG();
    
    sendstate = SENDSTATE_BREAKSYNC;
    
    TXSTAbits.SENDB = 1;
    TXREG = 0x00;
    TXREG = LIN_SYNC;
}

/**
 * @brief Send data only
 * @param data Data to send
 * @param length Length of Data
 * @param source Source (see REPORT_SOURCE_*)
 */
void lin_sendData(uint8_t * data, uint8_t length, uint8_t source) {

    if (length == 0) return;
    
    uint8_t i;
    
    if (sendsource != REPORT_SOURCE_USER) {
        sendsource = source;
    }
    finfo.reportid = sendsource | REPORT_TYPE_FRAME;
    senddatapos = 0;
    senddatalen = length;
    for (i = 0; i < length; i++) {
        senddata[i] = data[i];
    }   

    sendstate = SENDSTATE_DATA;
    TXREG = senddata[senddatapos++];    
}

/**
 * @brief Get current state of send module
 * @return State of send module
 */
uint8_t lin_getSendState(void) {
    return sendstate;
}

/**
 * @brief Determine if there is a character in receive buffer
 * @retval 0 No new character available
 * @retval 1 New character available
 */
uint8_t lin_chReceived(void) {
    return RCIF != 0;
}

/**
 * @brief Get protected identifier. Add parity bits to given id.
 * @param id Identifier
 * @return Protected identifier
 */
uint8_t lin_getPID(uint8_t id) {
    
    id = id & 0x3f;
    
    uint8_t p0 = ((id >> 0) + (id >> 1) + (id >> 2) + (id >> 4)) & 1;
    uint8_t p1 = ~((id >> 1) + (id >> 3) + (id >> 4) + (id >> 5)) & 1;    
    uint8_t parity = (uint8_t)(p0 | (p1<<1));
    
    return (uint8_t) (id | (parity << 6));
}

/**
 * @brief Get frame info
 * @return Frame info
 */
struct frameinfo * lin_getFrameinfo(void) {
    return &finfo;
}

/**
 * @brief Get current receive state
 * @return Receive state
 */
uint8_t lin_getReceiveState(void) {
    return receive_state;
}

/**
 * @brief Handle received frame. Give it to USB.
 */
void lin_receivedFrame(void) {
    usb_ep1_put((uint8_t*)&finfo, 16);
    receive_state = RECEIVESTATE_IDLE;
    framereceived = 1;
    timeoutticker = 0;    
}

/**
 * @brief Wait until current frame receiption is finished.
 */
void lin_waitFrameReceived(void) {
    framereceived = 0;
    while (!framereceived) {
        linslave_process();
    };
}

/**
 * @brief Determine if autobaud detection is set
 * @retval 0 No autobaud set
 * @retval 1 Autobaud is set
 */
uint8_t lin_autobaudDetectionIsSet(void) {
    return autobaud_activated && (!autobaud_suspended);
}

/**
 * @brief Cancel autobaud detection
 */
void lin_autobaudCancel(void) {
    
    BAUDCONbits.ABDEN = 0;
    BAUDCONbits.ABDOVF = 0;
    lin_setConfiguredSPBRG();
}

/**
 * @brief Suspend autobaud detection
 * @param suspended 1=suspend, 0=normal
 */
void lin_autobaudSuspend(uint8_t suspended) {
    autobaud_suspended = suspended;
}

/**
 * @brief Read one byte from LIN bus
 * @return Byte read from LIN bus
 */
uint8_t lin_readByte(void) {
    
    
    if (BAUDCONbits.ABDOVF) {
        lin_autobaudCancel();
        
        statusreport_setErrorflag(ERRORFLAG_ABAUD_OVERFLOW);
        
        finfo.length = 0;        
        receive_state = RECEIVESTATE_IDLE;
    }
    
    if (RCSTAbits.FERR) {
        led_flash();
        
        if (receive_state != RECEIVESTATE_IDLE) {
            // we are already in receive mode, finish reception      
            lin_receivedFrame();
        }        

        finfo.length = 0;        
        receive_state = RECEIVESTATE_IDLE;
        
        // Autobaud detection
        if (lin_autobaudDetectionIsSet()) {
            receive_state = RECEIVESTATE_ABAUD_START;
        }        
    }
    
    unsigned char ch = RCREG;
    
    if (finfo.length < RECEIVEBUFFER_SIZE) {
        finfo.data[finfo.length++] = ch;
        
        switch (receive_state) {
            
            case RECEIVESTATE_IDLE:
                
                if ((finfo.length == 2) && (finfo.data[0] == 0x00) && (finfo.data[1] == LIN_SYNC)) {
                    receive_state = RECEIVESTATE_PID;
                    finfo.abaud = SPBRG;
                    finfo.length = 0;
                    finfo.reportid = REPORT_SOURCE_COMMON | REPORT_TYPE_FRAME;
                    finfo.timestamp = clock_getMSTicker();                    
                }
                break;
                
            case RECEIVESTATE_ABAUD_START:
                
                receive_state = RECEIVESTATE_ABAUD_SYNC;
                BAUDCONbits.ABDEN = 1;
                break;
                
            case RECEIVESTATE_ABAUD_SYNC:
                
                    receive_state = RECEIVESTATE_PID;
                    finfo.abaud = SPBRG;
                    finfo.length = 0;
                    finfo.pid = 0;
                    finfo.reportid = REPORT_SOURCE_COMMON | REPORT_TYPE_FRAME;                    
                    finfo.timestamp = clock_getMSTicker();
                    
                    // sometimes auto baud detection skips the sync byte. if so, go directly to next state
                    if (ch != lin_getPID(ch))
                        break;
                
            case RECEIVESTATE_PID:
                
                finfo.pid = ch;
                finfo.length = 0;
                receive_state = RECEIVESTATE_DATA;
                
                // check PID
                if (ch != lin_getPID(ch)) {
                    
                    // invalid PID
                    receive_state = RECEIVESTATE_IDLE;
                    
                } else if ((sendstate == SENDSTATE_IDLE) || (senddatalen == 0)) {
                    
                    // trigger slave task
                    linslave_triggerIDProcessing(finfo.pid);
                }
                
                break;
                
            case RECEIVESTATE_DATA:
                
                if (finfo.length == RECEIVEBUFFER_SIZE) {
                    // reception finished, send it out
                    lin_receivedFrame();
                }
                break;
        }        
    }
    
    if (receive_state != RECEIVESTATE_IDLE) {
        timeoutticker = timeoutpreset;
    }
        
    return 0;
}

/**
 * @brief Do module processing
 */
void lin_process(void) {
    
    lin_processSend();    
}

/**
 * @brief Timmer interrupt function
 */
inline void lin_timeout_isr(void) {
    
    if (timeoutticker) {        
        timeoutticker--;
        
        if (timeoutticker == 0) {
            
            if (receive_state == RECEIVESTATE_ABAUD_SYNC) {
                lin_autobaudCancel();
            } else {            
                lin_receivedFrame();
            }
        }
        
    }    
}